package CTL;

import CTL.Annotate.*;
import CTL.Comm.*;
import CTL.Types.*;
import ReflWrap.*;

import gnu.getopt.*;
import java.io.*;
import java.lang.reflect.*;

/** Process group (currently, only groups with 2 members are supported) */
public class Group implements Runnable
{
	/** Logical ID of the process running on localhost */
	private int myID;
	/** Number of hosts in this group */
	private int members;
	/** ObjectID */
	private long objID;
	/** Array of GroupInfos for all members */
	protected GroupInfo[] grp;
	/** Whether the group is terminated */
	private boolean terminated = true;
	/** Environment of this group member */
	private Env env = null;

	/** Receive a termination request
	 * @param pid PeerID of the terminating process
	 * @param msg Reason for termination
	 */
	@builtin @sFID(3) public void recvTermination (PeerID pid, String msg)
	{
		env.log.msg(Logger.INFO, "recv TERM: "+pid+" - "+msg);
		System.exit(0);
	}

	/** Receive an exception
	 * @param msg Reason for the exception
	 */
	@builtin @sFID(4) public static void recvException (String msg)
	{
		// TODO: continue or terminate with recvTermination()
		throw new RuntimeException(msg);
	}

	/** Empty constructor */
	@cdont public Group (Env myEnv)
	{
		env = (myEnv == null) ? Env.newEnv() : myEnv;
		env.grp = this;

		String fe = System.getProperty("file.encoding");
		if (!fe.equals("ISO-8859-1"))
		{
			System.out.println("BUG: Bad file encoding '"+fe+"'.");
			System.exit(1);
		}

		Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
	}

	// Client which starts a remote service via SSH
	@cdont public Group (String host, int port, int myID, int members,
		int link, int myPort, Env myEnv) throws IOException, CTLException
	{
		this(myEnv);
	
		Env.newLogger("/tmp/client-"+myID);
		objID = env.map.regObj(this);
		env.log.msg(Logger.DBG1, "Registered as "+objID+".");
	
		env.comm = new Communicator[members];

		switch (link)
		{
			case Location.TCP:
				env.comm[0] = new TCPCommunicator();
				env.comm[1] = new TCPCommunicator(host, port, myPort);
				break;
			default:
				throw new CTLException("Linkage not implemented.");
		}
		
		env.comm[1].listen();

		updateGroup(myID, members);
		grp[myID] = new GroupInfo(env.comm[1].pid(), objID);
	}

	// Communication in daemon-mode
	@cdont public Group (String host, int port, int myID, int members, 
		int link, Env myEnv) throws IOException, 
		ClassNotFoundException, IllegalAccessException, 
		InstantiationException, InvocationTargetException, CTLException
	{
		this(myEnv);
		
		if (myEnv == null)
		{
			Env.newLogger("/tmp/service");
			objID = env.map.regObj(this);
			env.log.msg(Logger.DBG1, "Registered as "+objID+".");
	
			env.comm = new Communicator[members];
			
			switch (link)
			{
				case Location.HTTP:
					env.comm[0] = new HTTPCommunicator();
					env.comm[1] = new HTTPCommunicator();
					((HTTPCommunicator)env.comm[0]).setPort(port);
					break;
				case Location.TCP:
					env.comm[0] = new TCPCommunicator(port);
					env.comm[1] = new TCPCommunicator(port);
					break;
				default:
					throw new CTLException("Linkage not implemented.");
			}
			
			// FIXME: 0 or 1 as groupID
			env.comm[1].listen();
			updateGroup(myID, members);
			
			if (link == Location.TCP)
			{
				((TCPCommunicator)env.comm[1]).setHost(new IPaddr(host));
				System.out.println(env.comm[1].pid().port());
			}
			
			grp[myID] = new GroupInfo(env.comm[1].pid(), objID);
			env.daemon = true;
		}
		else
		{
			Env.newLogger("/tmp/client-"+myID);
			objID = env.map.regObj(this);
			env.log.msg(Logger.DBG1, "Registered as "+objID+".");

			env.comm = new Communicator[members];
			
			switch (link)
			{
				case Location.HTTP:
					env.comm[0] = new HTTPCommunicator();
					env.comm[1] = new HTTPCommunicator();
					break;
				case Location.TCP:
				case Location.DMN:
					env.comm[0] = new TCPCommunicator();
					env.comm[1] = new TCPCommunicator();
					break;
				default:
					throw new CTLException("Linkage not implemented.");
			}
				
			env.comm[0].listen();
			env.comm[0].connect(new PeerID(host, port));
			readVersion(env.comm[0]);
			
			updateGroup(myID, members);
			grp[0] = new GroupInfo(new PeerID(host, port), 0);
			grp[1] = new GroupInfo(env.comm[1].pid(), objID);

			Remote.writeDAT(env.comm[0], null, new IPaddr(Remote.magic, 0),
				env.comm[1].pid().port());
		}
	}

	// Remote service, started via SSH
	@cdont public Group (String[] args) throws IOException, CTLException, 
		IllegalAccessException, InvocationTargetException,
		ClassNotFoundException, InstantiationException
	{
		this((Env)null);
		
		Env.newLogger("/tmp/service");
		objID = env.map.regObj(this);
		env.log.msg(Logger.DBG1, "Registered as "+objID+".");

		String proto = "file";
		IPaddr host = new IPaddr();
		int port = 0, myID = 0, members = 1;
		
		Getopt g = new Getopt("Example.Server", args, "D:C:vL:S:f:h:l:cp:P:");
		int c;
		
		while ((c = g.getopt()) != -1)
		{
			// TODO: Implement missing options.
			switch (c)
			{
				case 'D': // Attach debugger
					env.debugger = RUtil.optPeerID(g);
					break;
				case 'C': // Commandline options for the real application
					//arg = g.getOptArg();
					break;
				case 'v': // Verbose
					break;
				case 'L': // Specify port
					port = RUtil.optInt(g);
					break;
				case 'p': // Specify logID
					myID = RUtil.optInt(g);
					break;
				case 'S': // Specify group size
					members = RUtil.optInt(g);
					break;
				case 'f': // Specify logfile
					//arg = g.getOptArg();
					break;
				case 'h': // Specify host
					host = new IPaddr(RUtil.optString(g));
					break;
				case 'l': // Specify linkage
					proto = RUtil.optString(g);
					break;
				case 'P': // Specify debugger port
					env.debugID = RUtil.optInt(g);
					break;
				case 'c': // Cypher
					break;
				case 'x': // Run in terminal (default: xterm -e)
					break;
				case 'g': // Start in debugger (default: gdb; implies -x)
					break;
				case 'd': // Change working directory to this
					break;
				case 's': // <remote shell> at <port> (default: ssh, port 22)
					break;
				case 'R': // Set remote port
					break;
					
				case '?':
					env.log.msg(Logger.ERR, "Unknown option "+(char)g.getOptopt()+
						" specified.");
					break;
				default:
					env.log.msg(Logger.ERR, "getopt returned "+c);
					break;
			}
		}
		
		env.log.msg(Logger.DBG, 
			proto+"://"+host+":"+port+" - "+myID+"/"+members);

		env.comm = new Communicator[members];

		if (proto.equals("tcp"))
		{
			env.comm[0] = new TCPCommunicator();
			env.comm[1] = new TCPCommunicator();
		}
		
		env.comm[0].listen();
		PeerID myPID = env.comm[0].pid();
		env.log.msg(Logger.DBG2, "My PID: "+myPID);
		env.comm[0].connect(new PeerID(host, port));
		readVersion(env.comm[0]);

		updateGroup(myID, members);
		grp[1] = new GroupInfo(myPID, objID);
		env.log.msg(Logger.DBG1, "Sending handshake to "+host+":"+port);
		HS_send();
	}

	/** Update group information
	 * @param myID Logical ID
	 * @param members Number of group members
	 */
	@dont private void updateGroup (int myID, int members)
	{
		this.myID = myID;
		this.members = members;
		if (members>0)
			grp = new GroupInfo[members];
	}
	
	/** Print information about the group */
	@dont public void printInfo ()
	{
		env.log.msg(Logger.DBG1, "I am "+toString()+" on "+
			grp[myID].toString());
	}

	/** Read version string
	 * @param comm Communicator
	 */
	@dont public void readVersion (Communicator comm) throws IOException,
		CTLException
	{
		byte[] buf = comm.recv(19);
		env.log.msg(Logger.INFO, new String(buf));
	}
	
	/** Write version string
	 * @param comm Communicator
	 */
	@dont public void writeVersion (Communicator comm) throws IOException
	{
		String str = "CTL v0.9 Java      ";
		comm.send(str.getBytes(), false);
	}

	/** Start the communication thread */
	@dont public void run ()
	{
		if (env.daemon)
		{
			try
			{
				env.comm[1].accept();
				writeVersion(env.comm[1]);
				Header head = (Header)Remote.readDAT(env.comm[1], "HEAD");
				IPaddr magic = new IPaddr(Remote.magic, 0);
				if (!head.pid().host().equals(magic))
				{
					env.log.msg(Logger.ERR, 
						"Magic numbers do not match.");
					throw new RuntimeException("Magic numbers do not match.");
				}
			} 
			catch (Exception e) 
			{
				RUtil.except(e);
			}
		}
		recv();
	}

	/** Spawns a connection handler as long as the group is active */
	@dont protected void recv ()
	{
		terminated = false;
		while (!terminated)
			new ConnectHandler(env).run();
	}

	/** Receive a handshake */
	@dont protected void HS_recv () throws IOException, 
		ClassNotFoundException, InstantiationException, 
		InvocationTargetException, IllegalAccessException,
		CTLException
	{
		BuffyIn in = new BuffyIn(env.comm[1].recv());
		in.i.serialRead(Header.class);
		
		//System.out.println("My GroupInfo: "+grp[0]);
		Remote.writeHS(grp[0], env.comm[1]);
		grp[1] = Remote.readHS(env.comm[1]);
		//System.out.println("GroupInfo of the service: "+grp[1]);
		env.log.msg(Logger.INFO, "Handshake received.");
	}
		
	/** Send a handshake */
	@dont protected void HS_send () throws IOException, 
		IllegalAccessException, InvocationTargetException,
		ClassNotFoundException, InstantiationException, 
		CTLException
	{
		Header head = new Header(0, Remote.DAT, new IPaddr(Remote.magic, 0),
			env.comm[0].pid().port(), myID());
		env.log.msg(Logger.DBG2, "HS_send(): "+head);
		BuffyOut out = new BuffyOut();
		out.o.serialWrite(head);
		out.o.flush();
		env.comm[0].send(out.getBytes());
		
		env.log.msg(Logger.INFO, "Sending handshake.");
		grp[0] = Remote.readHS(env.comm[0]);
		env.log.msg(Logger.DBG2, "Client's GroupInfo: "+grp[0]);
		Remote.writeHS(grp[1], env.comm[0]);
		env.log.msg(Logger.DBG2, "My GroupInfo: "+grp[1]);
		env.log.msg(Logger.INFO, "Handshake sent.");
	}

	/** Check if this group is active
	 * @return True if active, false otherwise
	 */
	@dont public boolean running ()
	{
		return !terminated;
	}

	/** Terminate the group */
	@dont public void terminate ()
	{
		terminated = true;
	}

	/** Retrieve local Logical ID
	 * @return Logical ID
	 */
	@dont public int myID ()
	{
		return myID;
	}

	/** Retrieve number of group members
	 * @return Number of members
	 */
	@dont public int members ()
	{
		return members;
	}

	/** Retrieve a string representation of this object
	 * @return String
	 */
	@dont public String toString ()
	{
		return myID+"/"+members;
	}

	/** Retrieve group information of a certain member */
	@dont public GroupInfo groupInfo (int id)
	{
		return grp[id];
	}

	/** Retrieve group information of the local member */
	@dont public GroupInfo myInfo ()
	{
		return grp[myID];
	}
}
