package CTL.Types;

import CTL.Serialize.*;
import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;

/** IP address storage and serialization for IPv4 and IPv6 */
public class IPaddr implements Writable
{
	/** 8 x 16 Bit = 128 Bit, which is the length of an IPv6 address */
	private short[] IP = new short[8];

	/** Retrieve the raw short array 
	 * @return Array
	 */
	public short[] rawIP ()
	{
		return IP;
	}

	/** Serial read function
	 * @param in Input stream
	 */
	public void read (SerialIn in) throws IOException, ClassNotFoundException, 
		InstantiationException, IllegalAccessException, InvocationTargetException
	{
		//System.out.println("DBG :: IPaddr.read() :: begin");
		Object[] tmp = in.__readArray(Short.TYPE, null, 8);
		for (int i=0;i<tmp.length;i++)
			IP[i]=((Short)tmp[i]).shortValue();
		//System.out.println("DBG :: IPaddr.read() :: end");
	}
	
	/** Serial write function
	 * @param out Output stream
	 */
	public void write (SerialOut out) throws IOException, IllegalAccessException,
		InvocationTargetException
	{
		out.__writeArray(IP);
	}

	/** Create an IPaddr from the Java InetAddress type
	 * @param in InetAddress
	 */
	public IPaddr (InetAddress in)
	{
		IP = getShortArray(in);
	}

	/** Create an IPaddr from an 'unsigned int32' 
	 * @param ip IP address
	 */
	public IPaddr (int ip) throws UnknownHostException
	{
		this(setUint32(ip));
	}

	/** Create an IPaddr from a hexadecimal string 
	 * @param str Hexadecimal number in a string
	 * @param i Dummy parameter
	 */
	public IPaddr (String str, int i)
	{
		if (str == null)
			return;
		StringTokenizer token = new StringTokenizer(str, ":");
		int m=0;
		while (token.hasMoreTokens())
		{
			String s = token.nextToken();
			IP[m]=0;
			for (int n=0;n<s.length();n++)
			{
				int c = (int)s.charAt(n);
				c-=48;
				if (c>9) c-=7;
				IP[m]+=c*Math.pow(16, 3-n);					
			}
			m++;
		}
	}

	/** Create an IPaddr from a hostname
	 * @param host Hostname
	 */
	public IPaddr (String host) throws UnknownHostException
	{
		this(InetAddress.getByName(host));
	}

	/** Create an IPaddr for localhost */
	public IPaddr () throws UnknownHostException
	{
		// Workaround for broken getLocalHost() behaviour
		try
		{
			IP = getShortArray(InetAddress.getLocalHost());
		}
		catch (UnknownHostException e)
		{
			IP = getShortArray(InetAddress.getByName("127.0.0.1"));
		}
	}

	/** Check if two IPaddrs are equal
	 * @param ip IPaddr to compare to
	 * @return True if equal, false if not
	 */
	public boolean equals (IPaddr ip)
	{
		if (ip==null)
			return ((toString().equals("")) ? true : false);
		if (toString().equals(ip.toString()))
			return true;
		return false;
	}

	/** Check if two InetAddresses are equal
	 * @param ip InetAddress to compare to
	 * @return True if equal, false if not
	 */
	public boolean equals (InetAddress ip)
	{
		short[] test = getShortArray(ip);

		for (int i=0;i<8;i++)
			if (test[i]!=IP[i])
				return false;
		return true;
	}

	/** Convert this to an integer
	 * @return Integer
	 */
	public int getUint32 () throws UnknownHostException
	{
		return getUint32(setShortArray(IP));
	}
	
	/** Convert ip to an integer
	 * @param ip InetAddress to convert
	 * @return Integer
	 */
	public static int getUint32 (InetAddress ip)
	{
		if (ip==null)
			return -1;
		byte[] adr = ip.getAddress();
		short[] i = new short[4];
		for (int j=0;j<4;j++)
			i[j] = (short) ((adr[j]<0) ? (256+adr[j]) : adr[j]);
		return i[3]+(i[2]<<8)+(i[1]<<16)+(i[0]<<24);
	}

	/** Convert an integer to InetAddress
	 * @param ip Integer
	 * @return InetAddress
	 */
	public static InetAddress setUint32 (int ip) throws UnknownHostException
	{
		StringBuffer buf = new StringBuffer();
		for (int i=32; i<=56;i+=8)
		{
			int t = ip<<i>>56;
			buf.append((t<0 ? 256+t : t)+(i==56 ? "" : "."));
		}
		InetAddress retval = InetAddress.getByName(buf.toString());
		return retval;
	}

	/** Retrieve a string representation of this object
	 * @return String
	 */
	public String toString ()
	{
		try
		{
			byte[] adr = setShortArray(IP).getAddress();
			short[] i = new short[4];
			for (int j=0;j<4;j++)
				i[j] = (short) ((adr[j]<0) ? (256+adr[j]) : adr[j]);
			return i[0]+"."+i[1]+"."+i[2]+"."+i[3];
		}
		catch (UnknownHostException e)
		{
			return "";
		}
	}

	/** Retrieve an InetAddress representation of this object
	 * @return InetAddress
	 */
	public InetAddress toInetAddress () throws UnknownHostException
	{
		return setShortArray(IP);
	}

	/** Convert a short array to InetAddress
	 * @param ip Short Array
	 * @return InetAddress
	 */
	public static InetAddress setShortArray (short[] ip) throws 
		UnknownHostException
	{
		if (ip.length!=8)
			return null;
		return setUint32((ip[6]<<16)+ip[7]);
	}

	/** Convert an InetAddress to a short array
	 * @param ip InetAddress
	 * @return Short array
	 */
	public static short[] getShortArray (InetAddress ip) 
	{
		int ip2 = getUint32(ip);
		short[] retval = new short[8];
		retval[6]=(short)(ip2>>16);
		retval[7]=(short)(ip2<<16>>16);
		return retval;
	}
}
