package Debug;

import CTL.*;
import CTL.Comm.*;
import CTL.Streams.*;
import CTL.Types.*;
import NCLib.*;

import java.io.*;
import java.net.*;
import java.util.*;

/** Very basic structure of a CTL debugger */
public class DummyDebugger
{
	private final static boolean manual = false;
	private static int recv = 0;
	
	public static void main (String[] a)
	{
		Env.newLogger("/tmp/debugger");

		Sink s = new Sink();
		s.start();

		if (manual)
		{
			ManualContinue mc = new ManualContinue(s);
			mc.start();
		}

		// El-cheapo identification of Debugger's hostname
		String myHost = "localhost";
		if (new File("locs.txt").exists())
		{
			LinkedList<Location> locs = Location.parseFile("locs.txt");
			if (locs != null)
			{
				String hisHost = locs.get(0).host();
				myHost = TCPCommunicator.determineIP(hisHost, 22).toString();
			}
		}

		String myProg = "Example.Client3"; // Specified by the user
		String cmd = "xterm -hold -e env CTL_DEBUGGER="+myHost+":"+s.getPort()+
			" java -Dfile.encoding=ISO-8859-1 "+myProg; // For C just use myProg
		try
		{
			System.out.println(NCLib.Process.exec(cmd));
		}
		catch (Exception e)
		{
			System.out.println("Could not start process:");
			e.printStackTrace();
			System.exit(11);
		}

		System.out.println("Received "+recv+" bytes.");
	}

	private static void sendContinuous (Sink sink, Socket sock)
	{
		try
		{
			Header head = new Header(0, Remote.CTRL, sink.getPID(), -1);
			OutputStream out0 = sock.getOutputStream();
			OOStream out1 = new OOStream(out0);
			out1.serialWrite(head);
			out1.flush();
		} catch (Exception e) {}
	}

	private static class ManualContinue extends Thread
	{
		private Sink sink;

		public ManualContinue (Sink s)
		{
			this.sink = s;
		}

		public void run ()
		{
			while (true)
			{
				try
				{
					int c = System.in.read();
				}
				catch (Exception e)
				{
					e.printStackTrace();
					System.exit(3);
				}

				for (Socket sock : sink.getDudes())
					sendContinuous(sink, sock);
			}
		}
	}

	private static class Sink extends Thread
	{
		private int port = -1;
		private ServerSocket ssock = null;
		private LinkedList<Socket> dudes = new LinkedList<Socket>();
		
		public Sink ()
		{
			try
			{
				ssock = new ServerSocket(0, 25);
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
			port = ssock.getLocalPort();
		}

		public void run ()
		{
			Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
			
			if (ssock == null)
				return;
			
			int i = 0;
			while (true)
			{
				try
				{
					Socket sock = ssock.accept();
					System.out.println("New connection from "+sock.getRemoteSocketAddress());
					// TODO: Remove sockets after end of communication
					dudes.add(sock);
					Handler handle = new Handler(sock, i, this);
					handle.start();
					i++;
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
		}

		public int getPort ()
		{
			return port;
		}

		public PeerID getPID () throws UnknownHostException
		{
			return new PeerID("localhost", port);
		}

		public LinkedList<Socket> getDudes ()
		{
			return dudes;
		}
	}

	/** This is where it gets interesting. The Handler class should pass the
	 * data it received to the remote object dummies. Most of that can be copied
	 * from the CTL4j. The other extension point is the CodeGen package, which
	 * needs a generator for the logging CIs. On top of that, there's a GUI
	 * which visualizes the connections between remote objects and their current
	 * state, as well as all the method calls. For this, DotClass can be a
	 * starting point, it shows how to easily get information about Java
	 * classes.
	 */
	private static class Handler extends Thread
	{
		private int id = -1;
		private Sink sink = null;
		private Socket sock = null;

		public Handler (Socket sock, int id, Sink sink)
		{
			this.id = id;
			this.sink = sink;
			this.sock = sock;
		}

		public void run ()
		{
			Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
			
			try
			{
				BufferedInputStream read = new BufferedInputStream(sock.getInputStream());
				OIStream in = new OIStream(read);
				while (sock.isConnected())
				{
					Header head = in.serialRead(Header.class);
					int size = (int)head.getSize();
					System.out.println(id+": "+head);
					DummyDebugger.recv += head.size();
					
					if (size > 64*1024*1024)
						return;

					if (size > 0)
					{
						byte[] buf = new byte[size];
						in.read(buf);
						DummyDebugger.recv += size;
					}

					if (!manual)
						sendContinuous(sink, sock);
				}
				sock.close();
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}
	}
}	
